// Copyright (C) Microsoft Corporation. All Rights Reserved.
// This code released under the terms of the Microsoft Public License
// (Ms-PL, http://opensource.org/licenses/ms-pl.html).

using System;
using System.Windows;
using System.Windows.Controls;

namespace Delay
{
    /// <summary>
    /// Abstract class that handles the core functionality of a "best fit" Panel.
    /// </summary>
    public abstract class BestFitPanel : Panel
    {
        /// <summary>
        /// Measures all children at the "best fit" size.
        /// </summary>
        /// <param name="availableSize">Available size.</param>
        /// <returns>Desired size.</returns>
        protected override Size MeasureOverride(Size availableSize)
        {
            var infiniteSize = double.IsInfinity(availableSize.Width) || double.IsInfinity(availableSize.Height);
            var childMeasureSize = availableSize;
            if (!infiniteSize)
            {
                // Calculate best fit based on available size
                var bestFitInfo = CalculateBestFit(availableSize, Children.Count);
                childMeasureSize = new Size(
                    Math.Round(availableSize.Width / bestFitInfo.Columns),
                    Math.Round(availableSize.Height / bestFitInfo.Rows));
            }
            var maximumDesiredWidth = 0.0;
            var maximumDesiredHeight = 0.0;
            foreach (UIElement child in Children)
            {
                // Always measure all children
                child.Measure(childMeasureSize);
                maximumDesiredWidth = Math.Max(child.DesiredSize.Width, maximumDesiredWidth);
                maximumDesiredHeight = Math.Max(child.DesiredSize.Height, maximumDesiredHeight);
            }
            return infiniteSize ?
                new Size(maximumDesiredWidth, maximumDesiredHeight) : // Return largest width/height instead of infinity
                availableSize; // Fill the available size
        }

        /// <summary>
        /// Arranges all children at the "best fit" size.
        /// </summary>
        /// <param name="finalSize">Final size.</param>
        /// <returns>Render size.</returns>
        protected override Size ArrangeOverride(Size finalSize)
        {
            var bestFitInfo = CalculateBestFit(finalSize, Children.Count);
            var row = 0;
            var column = 0;
            foreach (UIElement child in Children)
            {
                // Calculate both sides to avoid accumulating rounding errors with fractional
                // width/height. Round all values for optimal screen alignment.
                var x1 = Math.Round((column * finalSize.Width) / bestFitInfo.Columns);
                var y1 = Math.Round((row * finalSize.Height) / bestFitInfo.Rows);
                var x2 = Math.Round(((column + 1) * finalSize.Width) / bestFitInfo.Columns);
                var y2 = Math.Round(((row + 1) * finalSize.Height) / bestFitInfo.Rows);
                var childArrangeRect = new Rect(x1, y1, x2 - x1, y2 - y1);
                child.Arrange(childArrangeRect);
                column++;
                if (column == bestFitInfo.Columns)
                {
                    column = 0;
                    row++;
                }
            }
            return finalSize;
        }

        /// <summary>
        /// Calculates the "best fit" based on the specified size and element count.
        /// </summary>
        /// <param name="panelSize">Size of Panel.</param>
        /// <param name="elementCount">Count of elements.</param>
        /// <returns>BestFitInfo representing the "best fit".</returns>
        protected abstract BestFitInfo CalculateBestFit(Size panelSize, int elementCount);

        /// <summary>
        /// Class that represents the "best fit" settings.
        /// </summary>
        protected class BestFitInfo
        {
            /// <summary>
            /// Gets the number of columns.
            /// </summary>
            public int Columns { get; private set; }

            /// <summary>
            /// Gets the number of rows.
            /// </summary>
            public int Rows { get; private set; }

            /// <summary>
            /// Initializes a new instance of the BestFitInfo class.
            /// </summary>
            /// <param name="columns">Number of columns.</param>
            /// <param name="rows">Number of rows.</param>
            public BestFitInfo(int columns, int rows)
            {
                Columns = Math.Max(columns, 1);
                Rows = Math.Max(rows, 1);
            }
        }
    }

    /// <summary>
    /// Class that implements a "best fit" Panel that arranges so its elements have the
    /// biggest smallest dimension possible.
    /// </summary>
    public class MostBigPanel : BestFitPanel
    {
        /// <summary>
        /// Calculates the "best fit" based on the specified size and element count.
        /// </summary>
        /// <param name="panelSize">Size of Panel.</param>
        /// <param name="elementCount">Count of elements.</param>
        /// <returns>BestFitInfo representing the "best fit".</returns>
        protected override BestFitInfo CalculateBestFit(Size panelSize, int elementCount)
        {
            var bestColumns = 0;
            var bestRows = 0;
            var bestMinimumDimension = 0.0;
            for (var columns = 1; columns <= elementCount; columns++)
            {
                var rows = (elementCount + columns - 1) / columns;
                var itemWidth = panelSize.Width / columns;
                var itemHeight = panelSize.Height / rows;
                var minimumDimension = Math.Min(itemWidth, itemHeight);
                if (bestMinimumDimension < minimumDimension)
                {
                    bestMinimumDimension = minimumDimension;
                    bestColumns = columns;
                    bestRows = rows;
                }
            }
            return new BestFitInfo(bestColumns, bestRows);
        }
    }

    /// <summary>
    /// Class that implements a "best fit" Panel that arranges so its elements occupy
    /// the most space possible (i.e., leave the least blank space).
    /// </summary>
    public class MostFullPanel : BestFitPanel
    {
        /// <summary>
        /// Calculates the "best fit" based on the specified size and element count.
        /// </summary>
        /// <param name="panelSize">Size of Panel.</param>
        /// <param name="elementCount">Count of elements.</param>
        /// <returns>BestFitInfo representing the "best fit".</returns>
        protected override BestFitPanel.BestFitInfo CalculateBestFit(Size panelSize, int elementCount)
        {
            var bestColumns = 0;
            var bestRows = 0;
            var bestEmpty = int.MaxValue;
            var bestDelta = int.MaxValue;
            for (var columns = 1; columns <= elementCount; columns++)
            {
                var rows = (elementCount + columns - 1) / columns;
                var empty = (columns * rows) - elementCount;
                var delta = Math.Abs(rows - columns);
                if ((empty < bestEmpty) || (delta < bestDelta))
                {
                    bestEmpty = empty;
                    bestDelta = delta;
                    bestColumns = columns;
                    bestRows = rows;
                }
            }
            return new BestFitInfo(bestColumns, bestRows);
        }
    }

    /// <summary>
    /// Abstract class that handles the core functionality of a "best fit" Panel based on
    /// the angle of the diagonal of each item.
    /// </summary>
    public abstract class BestAnglePanel : BestFitPanel
    {
        /// <summary>
        /// Calculates the "best fit" based on the specified size and element count.
        /// </summary>
        /// <param name="panelSize">Size of Panel.</param>
        /// <param name="elementCount">Count of elements.</param>
        /// <returns>BestFitInfo representing the "best fit".</returns>
        protected override BestFitInfo CalculateBestFit(Size panelSize, int elementCount)
        {
            var idealAngle = GetIdealAngle(panelSize);
            var bestColumns = 0;
            var bestRows = 0;
            var bestAngle = 0.0;
            for (var columns = 1; columns <= elementCount; columns++)
            {
                var rows = (elementCount + columns - 1) / columns;
                var itemWidth = panelSize.Width / columns;
                var itemHeight = panelSize.Height / rows;
                var angle = Math.Atan2(itemHeight, itemWidth);
                if (Math.Abs(angle - idealAngle) < Math.Abs(bestAngle - idealAngle))
                {
                    bestAngle = angle;
                    bestColumns = columns;
                    bestRows = rows;
                }
            }
            return new BestFitInfo(bestColumns, bestRows);
        }

        /// <summary>
        /// Calculates the "ideal" angle for the diagonal of each item.
        /// </summary>
        /// <param name="panelSize">Size of the Panel.</param>
        /// <returns>Ideal angle.</returns>
        protected abstract double GetIdealAngle(Size panelSize);
    }

    /// <summary>
    /// Class that implements a "best fit" Panel that arranges so its elements are the
    /// most like a square.
    /// </summary>
    public class MostSquarePanel : BestAnglePanel
    {
        /// <summary>
        /// Calculates the "ideal" angle for the diagonal of each item.
        /// </summary>
        /// <param name="panelSize">Size of the Panel.</param>
        /// <returns>Ideal angle.</returns>
        protected override double GetIdealAngle(Size panelSize)
        {
            return Math.PI / 4; // 45 degree angle
        }
    }

    /// <summary>
    /// Class that implements a "best fit" Panel that arranges so its elements are the
    /// most like horizontally-oriented "golden rectangle".
    /// </summary>
    /// <remarks>
    /// See http://en.wikipedia.org/wiki/Golden_rectangle for more information.
    /// </remarks>
    public class MostGoldenPanel : BestAnglePanel
    {
        /// <summary>
        /// Calculates the angle of a "golden rectangle".
        /// </summary>
        private double _goldenAngle = Math.Atan2(1, (1 + Math.Sqrt(5)) / 2);

        /// <summary>
        /// Calculates the "ideal" angle for the diagonal of each item.
        /// </summary>
        /// <param name="panelSize">Size of the Panel.</param>
        /// <returns>Ideal angle.</returns>
        protected override double GetIdealAngle(Size panelSize)
        {
            return _goldenAngle;
        }
    }
}
